from collections.abc import Callable
from typing import TypeAlias


Filter: TypeAlias = Callable[[str], str]


def _version_major(version: str) -> str:
    return version.split(".")[0]


def _version_minor(version: str) -> str:
    return version.split(".")[1]


def _version_patch(version: str) -> str:
    return version.split(".")[2]


FILTERS: dict[str, Filter] = {
    "VersionMajor": _version_major,
    "VersionMinor": _version_minor,
    "VersionPatch": _version_patch,
}



from itertools import chain
from pathlib import Path
from typing import Annotated, Optional

from rich.console import Console
import typer

from nocto.env import get_env_variable, has_env_variable
from nocto.filters import FILTERS
from nocto.types import VariableOverrides
from nocto.variables import find_variables, replace_variables


app = typer.Typer()
stdout = Console()
stderr = Console(stderr=True)
File = Annotated[
    Path,
    typer.Argument(
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
        help="File in which to replace variables",
    ),
]
DotenvFile = Annotated[
    Optional[Path],  # noqa: UP007 - typer has problems with X | Y
    typer.Option(
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
        help="Optional .env file to use",
    ),
]
Dotenv = Annotated[bool, typer.Option(help="Use dotenv to load .env file")]
Vars = Annotated[
    Optional[list[str]],  # noqa: UP007 - typer has problems with X | Y
    typer.Option(help="Directly set variable value. E.g. FOO=BAR"),
]


def _process_variables_overrides(variables: Vars) -> VariableOverrides:
    if variables is None:
        return ()
    #                                                                       ↓ mypy does not understand maxsplit
    return tuple(tuple(var.split("=", maxsplit=1)) for var in variables)  # type: ignore[misc]


@app.command()
def test(file: File, dotenv: Dotenv = False, dotenv_file: DotenvFile = None, var: Vars = None) -> None:
    """
    Tests if local environment has all the variables required to replace Octopus-stype templated variables in `file`.
    """
    variables = find_variables(file)
    variable_overrides = _process_variables_overrides(var)
    missing_variables = sorted(
        {
            variable.name
            for variable in variables
            if not has_env_variable(variable.name, dotenv, dotenv_file, variable_overrides)
        }
    )
    if missing_variables:
        stderr.print(f"Missing environment variable{'s' if len(missing_variables) > 1 else ''}: {missing_variables}")
        raise typer.Exit(1)
    all_filters = chain.from_iterable(variable.filters for variable in variables)
    missing_filters = sorted({f for f in all_filters if f not in FILTERS})
    if missing_filters:
        stderr.print(f"Filter{'s' if len(missing_filters) > 1 else ''} not implemented: {missing_filters}")
        raise typer.Exit(1)


@app.command()
def replace(file: File, dotenv: Dotenv = False, dotenv_file: DotenvFile = None, var: Vars = None) -> None:
    """
    Replaces all Octopus-style template variables in `file` and writes it to temporary file.
    Returns path to temporary file.
    """
    test(file, dotenv, dotenv_file)
    variables = find_variables(file)
    variable_overrides = _process_variables_overrides(var)
    values = {
        variable: variable.process(get_env_variable(variable.name, dotenv, dotenv_file, variable_overrides))
        for variable in variables
    }
    output_path = replace_variables(file, values)
    stdout.print(output_path)


if __name__ == "__main__":
    app()



from collections.abc import Mapping
from functools import cache
import os
from pathlib import Path

from dotenv import dotenv_values

from nocto.types import VariableOverrides


@cache
def _env(dotenv: bool, dotenv_file: Path | None, variable_overrides: VariableOverrides) -> Mapping[str, str | None]:
    if not dotenv:
        return os.environ | dict(variable_overrides)
    return os.environ | dotenv_values(dotenv_file) | dict(variable_overrides)


def has_env_variable(name: str, dotenv: bool, dotenv_file: Path | None, variable_overrides: VariableOverrides) -> bool:
    return bool(_env(dotenv, dotenv_file, variable_overrides).get(name))


def get_env_variable(name: str, dotenv: bool, dotenv_file: Path | None, variable_overrides: VariableOverrides) -> str:
    value = _env(dotenv, dotenv_file, variable_overrides).get(name)
    if value is None:
        raise KeyError(f"Environment variable {name!r} not set")
    return value



from dataclasses import dataclass
from functools import cache, reduce
from pathlib import Path
import re
from tempfile import NamedTemporaryFile
from typing import Final

from nocto.filters import FILTERS


VARIABLES_REGEX: Final = re.compile(r"(?<!#)#{([^}]+)}", re.MULTILINE)


def _apply_filter(value: str, filter_: str) -> str:
    return FILTERS[filter_](value)


@dataclass(frozen=True, slots=True)
class Variable:
    name: str
    filters: tuple[str, ...]

    def process(self, value: str) -> str:
        """
        Applies `self.filters` in succession of the `value`
        """
        return reduce(_apply_filter, self.filters, value)


def _variable_from_string(variable: str) -> Variable:
    """
    Processes variables with filters - e.g. `Octopus.Release.Numbeer | VersionMajor`
    Into Variable(name='Octopus.Release.Number', filters=['VersionMajor'])
    """
    name, *filters = variable.split("|")
    return Variable(name.strip(), tuple(filter_.strip() for filter_ in filters))


@cache
def find_variables(file: Path) -> frozenset[Variable]:
    with file.open() as f:
        return frozenset(_variable_from_string(variable) for variable in VARIABLES_REGEX.findall(f.read()))


def replace_variables(file: Path, values: dict[Variable, str]) -> Path:
    def replace(match: re.Match) -> str:
        variable = _variable_from_string(match.group(1))
        return values[variable]

    with NamedTemporaryFile("w", delete=False) as temp_file, file.open() as in_file:
        replaced = VARIABLES_REGEX.sub(replace, in_file.read())
        temp_file.write(replaced)
        return Path(temp_file.name)
